import pygame
import random
import heapq

# Initialize pygame
pygame.init()

# Set up display
width, height = 800, 600
screen = pygame.display.set_mode((width, height))
pygame.display.set_caption("Snake Game")

# Colors
white = (255, 255, 255)
yellow = (255, 255, 102)
green = (0, 255, 0)
black = (0, 0, 0)
red = (213, 50, 80)

# Clock
clock = pygame.time.Clock()

# Snake settings
snake_block = 20
snake_speed = 15  # Default speed

# Fonts
font_style = pygame.font.SysFont("bahnschrift", 25)

# Message function
def message(msg, color, x, y):
    mesg = font_style.render(msg, True, color)
    screen.blit(mesg, [x, y])

# Place food (ensures it does not overlap the snake body)
def place_food(snake_body):
    while True:
        foodx = random.randint(0, (width - snake_block) // snake_block) * snake_block
        foody = random.randint(0, (height - snake_block) // snake_block) * snake_block
        if (foodx, foody) not in snake_body:
            return foodx, foody

# Optimized A* Algorithm
def a_star(start, goal, snake_body):
    directions = [(0, -snake_block), (0, snake_block), (-snake_block, 0), (snake_block, 0)]
    
    def heuristic(a, b):
        return abs(a[0] - b[0]) + abs(a[1] - b[1])

    open_list = []
    heapq.heappush(open_list, (0 + heuristic(start, goal), 0, start))  # (f, g, position)
    came_from = {}
    g_score = {start: 0}
    f_score = {start: heuristic(start, goal)}

    snake_body_set = set(snake_body)  # Convert to set for fast lookup

    while open_list:
        _, current_g, current = heapq.heappop(open_list)

        if current == goal:
            path = []
            while current in came_from:
                path.append(current)
                current = came_from[current]
            path.reverse()
            return path

        for direction in directions:
            neighbor = (current[0] + direction[0], current[1] + direction[1])
            if (0 <= neighbor[0] < width and 
                0 <= neighbor[1] < height and 
                neighbor not in snake_body_set and 
                neighbor not in g_score):
                tentative_g_score = current_g + 1
                g_score[neighbor] = tentative_g_score
                f_score[neighbor] = tentative_g_score + heuristic(neighbor, goal)
                heapq.heappush(open_list, (f_score[neighbor], tentative_g_score, neighbor))
                came_from[neighbor] = current

    return []  # No path found

# Main game loop
def game_loop():
    global snake_speed
    game_over = False
    game_close = False
    reason = ""  # Store the reason for game over

    # Snake initial position and direction
    x1, y1 = width // 2, height // 2

    # Snake body and initial length
    snake_list = [(x1, y1)]
    length_of_snake = 1

    # Place the initial food
    foodx, foody = place_food(snake_list)

    while not game_over:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                game_over = True
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_1:
                    snake_speed = 15
                elif event.key == pygame.K_2:
                    snake_speed = 30
                elif event.key == pygame.K_3:
                    snake_speed = 300

        while game_close:
            game_loop()
            screen.fill(black)
            message("Game Over!", red, width / 3, height / 3)
            message(f"Reason: {reason}", red, width / 3, height / 2)
            message("Press Q to Quit or C to Play Again", white, width / 6, height / 1.5)
            pygame.display.update()

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    game_over = True
                    game_close = False
                if event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_q:
                        game_over = True
                        game_close = False
                    if event.key == pygame.K_c:
                        game_loop()

        # Use A* to find the path to the food
        snake_head = snake_list[-1]
        path = a_star(snake_head, (foodx, foody), snake_list)

        # If a path exists, follow it step-by-step
        if path:
            next_step = path[0]
            x1, y1 = next_step
        else:
            reason = "No path to the food!"
            game_close = True

        # Update the snake's body
        snake_head = (x1, y1)
        if snake_head in snake_list:
            reason = "You collided with yourself!"
            game_close = True  # Snake ran into itself
        snake_list.append(snake_head)

        # Remove the tail if the snake hasn't grown
        if len(snake_list) > length_of_snake:
            snake_list.pop(0)

        # Check if food is eaten
        if x1 == foodx and y1 == foody:
            foodx, foody = place_food(snake_list)
            length_of_snake += 1

        # Drawing
        screen.fill(black)
        for segment in snake_list[:-1]:
            pygame.draw.rect(screen, green, [segment[0], segment[1], snake_block, snake_block])
        pygame.draw.rect(screen, red, [snake_list[-1][0], snake_list[-1][1], snake_block, snake_block])
        pygame.draw.rect(screen, yellow, [foodx, foody, snake_block, snake_block])

        # Display score and speed
        score = f"Score: {length_of_snake - 1}"
        message(score, white, 10, 10)
        speed_text = f"Speed: {snake_speed}"
        message(speed_text, white, 10, 40)

        pygame.display.update()
        clock.tick(snake_speed)

    pygame.quit()
    quit()

game_loop()
