import pygame
import time
import random
import math

# Initialize Pygame
pygame.init()

# Constants
SCREEN_WIDTH, SCREEN_HEIGHT = 800, 800
GRID_SIZE = 10
POINT_RADIUS = 5
BACKGROUND_COLOR = (30, 30, 30)
POINT_COLOR = (200, 200, 200)
PATH_COLOR = (255, 0, 0)
PATH_WIDTH = 4
TURRET_COLOR = (0, 255, 0)  # Green color for the turret body
SPIKE_COLOR = (255, 0, 0)  # Red color for the turret's spike (tip)
BULLET_COLOR = (255, 255, 0)  # Yellow color for the bullet
TURRET_MAX_RANGE = 300  # Max range of turret shooting
TURRET_ATTACK_SPEED = 2


gold = 50

# Calculate spacing
cell_width = SCREEN_WIDTH // GRID_SIZE
cell_height = SCREEN_HEIGHT // GRID_SIZE

# Predefined path (a complex path)
path_points = [
    (0, 1), (1, 1), (1, 7), (2, 7), (2, 9), (5, 9), (5, 8), (4, 8), (4, 5), (3, 5), (3, 4), (5, 4), (5, 2), (7, 2), (7, 6), (9, 6)
]

# Convert grid points to screen coordinates
def grid_to_screen(grid_point):
    col, row = grid_point
    x = col * cell_width + cell_width // 2
    y = row * cell_height + cell_height // 2
    return (x, y)

class Enemy:
    def __init__(self, hp, color, speed, gold_reward):
        self.hp = hp
        self.color = color
        self.speed = speed
        self.gold_reward = gold_reward  # Each enemy has a gold reward
        self.position_index = 0
        self.alive = True
        self.current_position = grid_to_screen(path_points[self.position_index])
        self.target_position = grid_to_screen(path_points[self.position_index + 1])
        self.dead_printed = False  # To ensure "-1" is printed only once

    def move(self):
        if self.position_index < len(path_points) - 1:
            start_x, start_y = self.current_position
            target_x, target_y = self.target_position
            dx = target_x - start_x
            dy = target_y - start_y
            distance = (dx ** 2 + dy ** 2) ** 0.5
            move_distance = self.speed

            if distance < move_distance:
                self.current_position = self.target_position
                self.position_index += 1
                if self.position_index < len(path_points) - 1:
                    self.target_position = grid_to_screen(path_points[self.position_index + 1])
            else:
                self.current_position = (start_x + dx * move_distance / distance, start_y + dy * move_distance / distance)

        if self.position_index == len(path_points) - 1 and self.alive:
            self.alive = False
            if not self.dead_printed:  # Only print once when the enemy dies
                print("-1")
                self.dead_printed = True  # Prevent further printing

    def draw(self, screen):
        if self.alive:
            pygame.draw.circle(screen, self.color, (int(self.current_position[0]), int(self.current_position[1])), POINT_RADIUS * 2)

# Bullet class (Homing missile behavior)
class Bullet:
    def __init__(self, start_pos, target_enemy, damage):
        self.position = start_pos
        self.target_enemy = target_enemy
        self.damage = damage
        self.speed = 1  # Bullet speed (2 times slower than the previous one)
        self.alive = True
        self.hit_enemies = []  # List to track enemies this bullet has already hit

        

    def move(self):
        if self.target_enemy.alive:
            # Calculate direction towards the enemy's current position
            target_x, target_y = self.target_enemy.current_position
            dx = target_x - self.position[0]
            dy = target_y - self.position[1]
            distance = math.sqrt(dx ** 2 + dy ** 2)

            # Normalize direction and move the bullet
            if distance != 0:
                dx /= distance
                dy /= distance
            self.position = (self.position[0] + dx * self.speed, self.position[1] + dy * self.speed)

            # If the bullet moves too far, delete it
            if distance < self.speed:
                self.alive = False
        else:
            self.alive = False

    def draw(self, screen):
        if self.alive:
            pygame.draw.circle(screen, BULLET_COLOR, (int(self.position[0]), int(self.position[1])), 5)

class Turret:
    def __init__(self, position):
        self.position = position
        self.color = TURRET_COLOR
        self.last_shoot_time = time.time()
        self.shooting_interval = TURRET_ATTACK_SPEED  # Time in seconds between shots
        self.bullets = []
        self.max_range = TURRET_MAX_RANGE  # Set the max shooting range

    def increase_attack_speed(self):  # Add a method to increase attack speed
        self.shooting_interval -= 0.2  # Decrease the time between shots
        print(f"Attack speed increased! New shooting interval: {self.shooting_interval}")


    def find_nearest_enemy(self, enemies):
        nearest_enemy = None
        min_distance = float('inf')
        for enemy in enemies:
            if enemy.alive:  # Ensure only alive enemies are considered
                dx = enemy.current_position[0] - self.position[0]
                dy = enemy.current_position[1] - self.position[1]
                distance = math.sqrt(dx ** 2 + dy ** 2)
                if distance < min_distance:
                    nearest_enemy = enemy
                    min_distance = distance
        return nearest_enemy

    def shoot(self, nearest_enemy):
        # Create a homing bullet that tracks the nearest enemy
        bullet = Bullet(start_pos=self.position, target_enemy=nearest_enemy, damage=1)
        self.bullets.append(bullet)
        print("Shot fired!")

    def update(self, enemies):
        # Find the nearest enemy and check if it is in range
        nearest_enemy = self.find_nearest_enemy(enemies)
        if nearest_enemy:
            dx = nearest_enemy.current_position[0] - self.position[0]
            dy = nearest_enemy.current_position[1] - self.position[1]
            distance = math.sqrt(dx ** 2 + dy ** 2)

            # Shoot only if the enemy is within range
            if distance <= self.max_range and (time.time() - self.last_shoot_time) >= self.shooting_interval:
                self.shoot(nearest_enemy)
                self.last_shoot_time = time.time()

        # Update and check bullets for collision with living enemies
        for bullet in self.bullets[:]:
            if bullet.alive:
                bullet.move()
                bullet.draw(screen)

                # Check if bullet hits a living enemy
                for enemy in enemies:
                    if enemy.alive and math.dist(bullet.position, enemy.current_position) < 10:  # Collision detection
                        if enemy not in bullet.hit_enemies:  # Only trigger collision once per enemy per bullet
                            bullet.hit_enemies.append(enemy)  # Track the enemy this bullet hit
                            enemy.hp -= bullet.damage
                            print("coll")
                            if enemy.hp <= 0:
                                enemy.alive = False
                                global gold
                                gold += enemy.gold_reward  # Add gold for the enemy's death
                            bullet.alive = False
                            break  # Exit the collision loop once the bullet hits an enemy

            else:
                self.bullets.remove(bullet)  # Remove bullet if it's not alive anymore

        # Rotate turret to face nearest enemy
        if nearest_enemy:
            dx = nearest_enemy.current_position[0] - self.position[0]
            dy = nearest_enemy.current_position[1] - self.position[1]
            angle = math.atan2(dy, dx)
            rotated_image = pygame.Surface((30, 30), pygame.SRCALPHA)
            
            # Draw the body of the turret (triangle)
            pygame.draw.polygon(rotated_image, self.color, [
                (15, 0), (30, 30), (0, 30)
            ])
            
            # Draw the spike of the turret in red (at the tip)
            pygame.draw.polygon(rotated_image, SPIKE_COLOR, [
                (15, -15), (17, 0), (13, 0)
            ])

            rotated_image = pygame.transform.rotate(rotated_image, math.degrees(angle))
            rotated_rect = rotated_image.get_rect(center=(self.position[0], self.position[1]))
            screen.blit(rotated_image, rotated_rect.topleft)

        # Draw the turret's range
        pygame.draw.circle(screen, (0, 0, 255), (int(self.position[0]), int(self.position[1])), self.max_range, 2)

    def draw(self, screen):
        # We don't need to draw a static turret here since it's handled in the update method
        pass


def open_shop(screen):
    # Draw the black background for the shop
    screen.fill((0, 0, 0))  # Black background for the shop
    
    # Draw the "Shop" title at the top center
    font = pygame.font.SysFont("Arial", 36)
    title_text = font.render("Shop", True, (255, 255, 255))
    screen.blit(title_text, (SCREEN_WIDTH // 2 - title_text.get_width() // 2, 50))
    
    # Draw the table with borders (5x2)
    table_width = 400
    table_height = 200
    table_top_left = (SCREEN_WIDTH // 2 - table_width // 2, SCREEN_HEIGHT // 2 - table_height // 2)
    cell_width = table_width // 5
    cell_height = table_height // 2

    # Draw table background
    pygame.draw.rect(screen, (255, 255, 255), (table_top_left[0], table_top_left[1], table_width, table_height))

    # Draw the grid lines for 5x2 table
    for i in range(6):  # 5 columns, so 6 lines
        pygame.draw.line(screen, (0, 0, 0), (table_top_left[0] + i * cell_width, table_top_left[1]), 
                         (table_top_left[0] + i * cell_width, table_top_left[1] + table_height), 2)

    for i in range(3):  # 2 rows, so 3 lines
        pygame.draw.line(screen, (0, 0, 0), (table_top_left[0], table_top_left[1] + i * cell_height), 
                         (table_top_left[0] + table_width, table_top_left[1] + i * cell_height), 2)
        
    item_data = [
        ("Att. sp", "0", (255, 0, 0), 'attack_speed'),
        ("Turr rng", "0", (0, 255, 0), 'turret_range'),
    ]

    items = []

    for row in range(2):
        for col in range(5):
            if row * 5 + col < len(item_data):  # Check if there's an item to display
                item_name, description, color, icon_type = item_data[row * 5 + col]
                
                # Render item name and description
                font_small = pygame.font.SysFont("Arial", 20)
                text = font_small.render(f"{item_name}: {description}", True, (255, 0, 0))
                
                # Position for the item description
                item_rect = pygame.Rect(table_top_left[0] + col * cell_width,
                                        table_top_left[1] + row * cell_height,
                                        cell_width,
                                        cell_height)
                screen.blit(text, (item_rect.centerx - text.get_width() // 2, item_rect.centery - text.get_height() // 2))
                
                # Draw item icons based on type
                if icon_type == 'attack_speed':
                    pygame.draw.circle(screen, color, (item_rect.centerx, item_rect.centery - 20), 15)  # Attack Speed Icon (circle)
                elif icon_type == 'turret_range':
                    pygame.draw.rect(screen, color, (item_rect.centerx - 15, item_rect.centery - 25, 30, 30))  # Turret Range Icon (square)

                items.append(item_rect)  # Store the rectangle for each item

    # Optional: Add some placeholder text for each item
    font_small = pygame.font.SysFont("Arial", 20)
    gold_text = font_small.render(f"Gold: {gold}", True, (255, 255, 0))  # Display current gold
    screen.blit(gold_text, (10, 10))

    return items

# Add a flag to track turret placement mode
placing_turret = False
turret_preview_pos = None

def handle_turret_placement():
    global placing_turret, turret_preview_pos
    mouse_x, mouse_y = pygame.mouse.get_pos()

    # Convert the mouse position to grid coordinates
    col = mouse_x // cell_width
    row = mouse_y // cell_height
    turret_preview_pos = grid_to_screen((col, row))

    # If the player left-clicks, place the turret
    if placing_turret and pygame.mouse.get_pressed()[0]:  # Left click
        # Place the turret at the grid position
        turret = Turret(position=turret_preview_pos)
        turrets.append(turret)  # Add to the turret list
        placing_turret = False  # End turret placement mode
        print(f"Turret placed at ({col}, {row})")

# Define buy functions for each item
def buy1():
    global gold  # Use the global gold variable
    if gold >= 50:  # Check if the player has enough gold
        gold -= 50
        for turret in turrets:
            turret.increase_attack_speed()
        print(f"Item 1 bought! You now have {gold} gold.")
        # Add other logic for buying item 1
    else:
        print("Not enough gold to buy Item 1.")

def buy2():
    global gold  # Use the global gold variable
    global shop_open
    global placing_turret
    if gold >= 50:  # Check if the player has enough gold
        gold -= 50
        shop_open = not shop_open
        placing_turret = True
        print("Placing turret...")
        print(f"Item 2 bought! You now have {gold} gold.")
        # Add other logic for buying item 1
    else:
        print("Not enough gold to buy Item 2.")
    

def buy3():
    print("Buying item 3")
    
    # Implement the logic for buying item 3
    pass

def buy4():
    print("Buying item 4")
    # Implement the logic for buying item 4
    pass

def buy5():
    print("Buying item 5")
    # Implement the logic for buying item 5
    pass

def buy6():
    print("Buying item 6")
    # Implement the logic for buying item 6
    pass

def buy7():
    print("Buying item 7")
    # Implement the logic for buying item 7
    pass

def buy8():
    print("Buying item 8")
    # Implement the logic for buying item 8
    pass

def buy9():
    print("Buying item 9")
    # Implement the logic for buying item 9
    pass

def buy10():
    print("Buying item 10")
    # Implement the logic for buying item 10
    pass


# Check for mouse click
def handle_click(items):
    mouse_x, mouse_y = pygame.mouse.get_pos()
    for idx, item_rect in enumerate(items):
        if item_rect.collidepoint(mouse_x, mouse_y):  # Check if the mouse click is inside the item rectangle
            print(f"Item {idx + 1} clicked!")  # Print the item number (1-based index)
            # Call the corresponding buy function based on the item index
            if idx == 0:
                buy1()
            elif idx == 1:
                buy2()
            elif idx == 2:
                buy3()
            elif idx == 3:
                buy4()
            elif idx == 4:
                buy5()
            elif idx == 5:
                buy6()
            elif idx == 6:
                buy7()
            elif idx == 7:
                buy8()
            elif idx == 8:
                buy9()
            elif idx == 9:
                buy10()
            break



# Create the screen
screen = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT))
pygame.display.set_caption("Tower Defense Grid")

# List to hold all enemies
enemies = []

# Timer for spawning enemies
last_spawn_time = time.time()
spawn_interval = random.randint(2, 5)  # Random time between 2 and 5 seconds

# Create the turret at position (3, 3)
# turret1 = Turret(position=grid_to_screen((3, 3)))
# turret2 = Turret(position=grid_to_screen((8, 1)))

# Game loop
running = True
turrets = []  # List to store placed turrets
shop_open = False  # Track whether the shop is open or closed

items = open_shop(screen)  # Render the shop and get the item rectangles

while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False
        if event.type == pygame.MOUSEBUTTONDOWN:
            if shop_open == True:
                handle_click(items)  # Handle clicks on the items
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_1:
                new_enemy = Enemy(hp=4, color=(255, 255, 0), speed=0.1, gold_reward = 10)  # Create a new enemy
                enemies.append(new_enemy)  # Add it to the list of enemies
                print("New enemy spawned!")
            if event.key == pygame.K_2:
                # Toggle the shop state (pause game and show shop)
                shop_open = not shop_open
                print("Shop Opened!" if shop_open else "Shop Closed!")

        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_3:  # Key 3 for placing a turret
                placing_turret = True
                print("Placing turret...")
        if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 3:  # Left-click
                handle_turret_placement()

        # If placing turret, show a preview
    if placing_turret and turret_preview_pos:
        pygame.draw.circle(screen, TURRET_COLOR, (int(turret_preview_pos[0]), int(turret_preview_pos[1])), 20)  # Preview



    # If the shop is open, pause the game logic
    if shop_open:
        open_shop(screen)
    else:
        # Fill the screen
        screen.fill(BACKGROUND_COLOR)

        # Draw the grid points
        for row in range(GRID_SIZE):
            for col in range(GRID_SIZE):
                x = col * cell_width + cell_width // 2
                y = row * cell_height + cell_height // 2
                pygame.draw.circle(screen, POINT_COLOR, (x, y), POINT_RADIUS)

        # Draw the path
        path_screen_points = [grid_to_screen(point) for point in path_points]
        pygame.draw.lines(screen, PATH_COLOR, False, path_screen_points, PATH_WIDTH)

        # Spawn new enemy if enough time has passed
        current_time = time.time()
        if current_time - last_spawn_time >= spawn_interval:
            # Create a new enemy with HP = 2
            new_enemy = Enemy(hp=1, color=(255, 0, 0), speed=0.1, gold_reward = 10)  # Speed is in pixels per frame
            enemies.append(new_enemy)
            last_spawn_time = current_time
            spawn_interval = random.randint(1, 13)/10  # Reset the spawn interval

        # Move and draw enemies
        for enemy in enemies[:]:
            if enemy.alive:
                enemy.move()
                enemy.draw(screen)
            else:
                enemies.remove(enemy)  # Remove dead enemies from the list

        for turret in turrets:
            turret.update(enemies)  # Update turrets and draw them
                

    # Update the screen
    pygame.display.update()
    

# Quit Pygame
pygame.quit()
